/******************************************************************************

            Copyright (c) 1996 - 2004  --  University of Washington

******************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../include/Cgen.h"
#include "../include/Dgen.h"
#include "../include/Privgen.h"
#include "../include/Stackgen.h"
#include "../include/buildstmt.h"
#include "../include/buildzplstmt.h"
#include "../include/dimension.h"
#include "../include/error.h"
#include "../include/expr.h"
#include "../include/global.h"
#include "../include/parsetree.h"
#include "../include/runtime.h"
#include "../include/stmtutil.h"
#include "../include/symboltable.h"
#include "../include/symmac.h"
#include "../include/symtab.h"
#include "../include/treemac.h"
#include "../include/typeinfo.h"


static void DeclareDynamicRegion(FILE* outfile, expr_t* reg) {
  fprintf(outfile, "_REG_DECL(, _reg_info, _region_fnc, ");
  gen_name(outfile, reg);
  fprintf(outfile, ");\n");
}


static void DeclareDynamicPrepRegion(FILE* outfile, expr_t *reg) {
  expr_t* basereg;
  expr_t* dir;

  basereg = T_OPLS(reg);
  dir = T_NEXT(T_OPLS(reg));
  if (T_IDENT(dir) && S_SETUP(T_IDENT(dir))==0) {
    force_pst_init(outfile, T_IDENT(dir));
  }
  MakeDRDeclarations(outfile, T_OPLS(reg));
  DeclareDynamicRegion(outfile, reg);
}


static void SetupDynamicRegionDims(FILE* outfile, datatype_t *pdt,
				   expr_t* reg) {
  int i;
  int numdims;
  int flooddim;
  int rgriddim;

  numdims = D_REG_NUM(pdt);
  fprintf(outfile, "  _NUMDIMS(");
  gen_expr(outfile, reg);
  fprintf(outfile, ") = %d;\n",numdims);
  fgenf(outfile, "  _REG_ARRLIST(%E) = NULL;\n", reg);
  for (i=0;i<numdims;i++) {
    flooddim = D_REG_DIM_TYPE(pdt, i) == DIM_FLOOD;
    rgriddim = D_REG_DIM_TYPE(pdt, i) == DIM_GRID;
    if (D_REG_DIM_TYPE(pdt, i) == DIM_INHERIT) {
      fprintf(outfile,"  _INHERIT_GLOB_BOUNDS(");
      gen_expr(outfile, reg);
      fprintf(outfile, ",%s.%s,%d);\n",RMS,RMS_REG,i);
    } else {
      fprintf(outfile,"  _SET_GLOB_BOUNDS");
      if (flooddim || rgriddim) {
	fprintf(outfile,"_FLOOD");
      }
      fprintf(outfile,"(");
      gen_expr(outfile, reg);
      fprintf(outfile, ",%d,",i);
      if (flooddim || rgriddim) {
	fprintf(outfile,"0,0,");
	if (flooddim) {
	  fprintf(outfile,"_DIM_FLOODED");
	} else {
	  fprintf(outfile,"_DIM_RGRID");
	}
      } else {
	INT_FATAL(NULL, "Not looking for dynamic dimensions in the right place");
      }
      fprintf(outfile,");\n");
    }
  }
}


static void SetupDynamicRegion(FILE* outfile, expr_t* reg,
			       statement_t* stmtls) {
  datatype_t *pdt;

  pdt = T_TYPEINFO(reg);

  fprintf(outfile, "NO COMPILE: WooHoo -- SetupDynamicRegion;\n");
  if (use_default_domain) {
    fprintf(outfile, "  _REG_DOM(");
    gen_expr(outfile, reg);
    fprintf(outfile, ") = &_DefaultDomain;\n");
    /*    fgenf(outfile, "  _DOM_REGISTER(&_DefaultDomain, %E);\n", reg);*/
  }
  else {
    fprintf(outfile, "  _REG_DOM(");
    gen_expr(outfile, reg);
    fprintf(outfile, ") = _REG_DOM(_ARR_DECL_REG(%s));\n",
	    S_IDENT(expr_find_ens_pst(stmtls_find_ens_expr(stmtls))));
    /*    fgenf(outfile, "  _DOM_REGISTER(_REG_DOM(_ARR_DECL_REG(%s)), %E);\n",
	  S_IDENT(expr_find_ens_pst(stmtls_find_ens_expr(stmtls))), reg);*/
  }

  SetupDynamicRegionDims(outfile, pdt, reg);

  fprintf(outfile,"_SetupDynamicRegion(");
  gen_expr(outfile, reg);
  fprintf(outfile, ");\n");

}


static void SetupDynamicPrepRegion(FILE* outfile, expr_t *reg, 
				   statement_t* stmtls) {
  expr_t *basereg;
  expr_t *dir;

  fprintf(outfile, "NO COMPILE: WooHoo -- SetupDynamicPrepRegion;\n");
  basereg = T_OPLS(reg);
  dir = T_NEXT(T_OPLS(reg));

  SetupDRs(outfile, basereg, stmtls);

  if (expr_dyn_reg(basereg)) {
    if (use_default_domain) {
      fprintf(outfile, "  _REG_DOM(");
      gen_expr(outfile,basereg);
      fprintf(outfile, ") = &_DefaultDomain;\n");
    /*    fgenf(outfile, "  _DOM_REGISTER(&_DefaultDomain, %E);\n", reg);*/
    }
    else {
      fprintf(outfile, "  _REG_DOM(");
      gen_expr(outfile, basereg);
      fprintf(outfile, ") = _REG_DOM(_ARR_DECL_REG(%s));\n",
	      S_IDENT(expr_find_ens_pst(stmtls_find_ens_expr(stmtls))));
    /*    fgenf(outfile, "  _DOM_REGISTER(_REG_DOM(_ARR_DECL_REG(%s)), %E);\n",
	  S_IDENT(expr_find_ens_pst(stmtls_find_ens_expr(stmtls))), reg);*/
    }

    SetupDynamicRegionDims(outfile, T_TYPEINFO(basereg),basereg);
  }
  fprintf(outfile,"_SetupDynamic");

  switch (T_SUBTYPE(reg)) {
  case OF_REGION:
    fprintf(outfile,"Of");
    break;
  case AT_REGION:
    fprintf(outfile,"At");
    break;
  case IN_REGION:
    fprintf(outfile,"In");
    break;
  case BY_REGION:
    fprintf(outfile,"By");
    break;
  default:
    INT_FATAL(NULL,"Unexpected region type in SetupDynamicPrepRegion");
  }

  fprintf(outfile,"Region(");
  gen_name(outfile, reg);
  fprintf(outfile, ",");
  gen_name(outfile, basereg);
  fprintf(outfile,",");
  gen_name(outfile,dir);
  fprintf(outfile,");\n");
}


void MakeDRDeclarations(FILE* outfile, expr_t *reg) {
  if (expr_is_qreg(reg)) {
    return;
  } else if (expr_contains_qreg(reg)) {
    switch (T_SUBTYPE(reg)) {
    case AT_REGION:
    case IN_REGION:
    case OF_REGION:
    case BY_REGION:
      DeclareDynamicRegion(outfile, reg);
      break;
    case INTERNAL_REGION:
      break;
    case DYNAMIC_REGION:
      INT_FATAL(NULL, "Should never get here");
      break;
    default:
      INT_FATAL(NULL, "Unexpected region type in push_region_scope(): %d",
		T_SUBTYPE(reg));
      break;
    }
  } else {
    if (T_SUBTYPE(reg) == DYNAMIC_REGION) {
      DeclareDynamicRegion(outfile, reg);
    } else if (T_IDENT(reg) == NULL && expr_prep_reg(reg)) {
      DeclareDynamicPrepRegion(outfile, reg);
    }
  }
}


static void SetupDynamicDirQuoteRegion(FILE* outfile, expr_t *reg) {
  int numdims;
  expr_t *dir;

  numdims = D_REG_NUM(T_TYPEINFO(reg));
  dir = T_NEXT(T_OPLS(reg));

  fprintf(outfile,"_AdjustRegionWith");
  switch (T_SUBTYPE(reg)) {
  case AT_REGION:
    fprintf(outfile,"At");
    break;
  case IN_REGION:
    fprintf(outfile,"In");
    break;
  case OF_REGION:
    fprintf(outfile,"Of");
    break;
  case BY_REGION:
    fprintf(outfile,"By");
    break;
  default:
    INT_FATAL(NULL,"BLC -- should never get here");
  }
  fprintf(outfile,"(");
  gen_name(outfile, reg);
  fprintf(outfile, ",%s.%s,",RMS,RMS_REG);

  mi_reg = pst_qreg[numdims];
  gen_expr(outfile,dir);
  mi_reg = NULL;

  fprintf(outfile,");\n");
}


void SetupDRs(FILE* outfile, expr_t* reg, statement_t* stmtls) {
  if (expr_is_qreg(reg)) {
    return;
  } else if (expr_contains_qreg(reg)) {
    switch(T_SUBTYPE(reg)) {
    case AT_REGION:
    case IN_REGION:
    case OF_REGION:
    case BY_REGION:
      SetupDynamicDirQuoteRegion(outfile, reg);
      break;
    case INTERNAL_REGION:
      break;
    case DYNAMIC_REGION:
      INT_FATAL(NULL, "Should never get here");
      break;
    default:
      INT_FATAL(NULL, "Unexpected region type in SetupDRs(): %d",
		T_SUBTYPE(reg));
      break;
    }
  } else {
    if (T_IDENT(reg) == NULL || (!S_SETUP(T_IDENT(reg)))) {
      if (T_TYPE(reg) == BIPREP) {
	/* This may be OK, but I haven't found need for it yet */
	/*	SetupDRs(outfile, T_OPLS(reg), stmtls);*/
	SetupDynamicPrepRegion(outfile, reg, stmtls);
      } else {
	if (T_IDENT(reg) != NULL) {
	  force_pst_init(outfile, T_IDENT(reg));
	  force_pst_finalize(outfile, T_IDENT(reg));
	  if (0) {
	    SetupDynamicRegion(outfile, reg, stmtls);
	  }
	}
      }
    }
  }
}


static void PushRegionMask(FILE* outfile, expr_t *reg, expr_t *maskexpr,
			   int maskbit, statement_t* stmtls) {
  symboltable_t *maskvar=NULL;

  if (maskexpr!=NULL) {
    maskvar = get_symtab_from_lvalue(maskexpr);
  }
  

  /* Create a new scope and save the old region stack */
  fprintf(outfile, "{  /* Push region scope */\n");
  
  /* save the old values of the region stack no matter what */
  /* save the old values of the mask stack no matter what */

  /* declare and initialize any dynamic regions as necessary */
  MakeDRDeclarations(outfile, reg);
  
  fprintf(outfile,"\n");

  /* set up any dynamic regions as necessary */
  SetupDRs(outfile, reg, stmtls);
  
  /* push the new values on the stack */
  
  fprintf(outfile,"%s.%s = ",RMS,RMS_REG);
  gen_name(outfile, reg);
  fprintf(outfile, ";\n");


  if (maskvar == NULL) {
    fprintf(outfile,"%s.%s = NULL;\n",RMS,RMS_MASK);
  } else {
    if (!expr_is_qmask(maskexpr)) {
      fprintf(outfile,"%s.%s = ",RMS,RMS_MASK);
      
      mask_no_access++;
      gen_expr(outfile, maskexpr);
      mask_no_access--;
      
      fprintf(outfile,";\n");
    }
    if (maskbit != -1) {
      fprintf(outfile,"%s.%s = %d;\n",RMS,RMS_WITH,maskbit);
    }
  }
  fprintf(outfile,"\n"); 
}


/*************************************************************
  Pushes the region and mask information in a region scope node
  onto the runtime "stack" implemented using lexical scoping.

  Ruth Anderson 5-27-93
  modified by Brad 1-13-96
************************************************************/

void push_region_scope(FILE* outfile, region_t *region_p, 
		       statement_t* stmtls) {
  /* Push the current region(s) onto the appropriate bucket of region stack*/
    
  if (region_p != NULL) {
    PushRegionMask(outfile, T_REGION_SYM(region_p),T_MASK_EXPR(region_p),
		   T_MASK_BIT(region_p), stmtls);
  }
}


/*************************************************************
  Pops the regions and masks associated with this region_scope_node
  off the region stack.  After this call, the RegionStack will be
  restored to its previous values at the time the regions associated
  with this region_scope_node were pushed onto the RegionStack.
  
  Ruth Anderson 5-27-93
  modified by Brad 1-13-96
************************************************************/

void pop_region_scope(FILE* outfile, region_t *region_p) {
  int numdims;
  expr_t * reg;
  symboltable_t * maskvar;

  if (region_p != NULL) {
    reg = T_REGION_SYM(region_p);
    if (T_MASK_EXPR(region_p)) {
      maskvar = get_symtab_from_lvalue(T_MASK_EXPR(region_p));
    } else {
      maskvar = NULL;
    }
    if (reg != NULL) {
      numdims = D_REG_NUM(T_TYPEINFO(reg));
    } else if (maskvar != NULL) {
      numdims = D_REG_NUM(T_TYPEINFO(T_TYPEINFO_REG(T_MASK_EXPR(region_p))));
    } else {
      INT_FATAL(NULL, "Region scope with no information!");
    }
  }
  
  fprintf(outfile, "} /* Pop region scope */ \n");

}

/***
 ***
 *** FUNCTION: genStackReg()
 ***           Prints out the "current" region or mask.
 ***
 *** AUTHOR: sungeun (5 February 1998)
 ***
 ***/

/* BLC -- number of dims no longer matters; use "1" arbitrarily */
void genStackReg(FILE* f) {
  priv_reg_access(f, buildqregexpr(0));
}

void OpenFloodReduceDynamicRegion(FILE* outfile, expr_t *reg) {
  fprintf(outfile,"{\n");
  MakeDRDeclarations(outfile, reg);
  SetupDRs(outfile, reg, NULL);
}


void CloseFloodReduceDynamicRegion(FILE* outfile, expr_t *reg) {
  fprintf(outfile,"}\n");
}
