/******************************************************************************

            Copyright (c) 1996 - 2004  --  University of Washington

******************************************************************************/


/***************************************************************\
* perregion.c is used to generate code on a per-region basis.   *
* It is chiefly responsible for dividing the regions amongst    *
* processors at runtime.                                        *
\***************************************************************/

#include <stdlib.h>
#include <string.h>
#include "../include/Mgen.h"
#include "../include/Privgen.h"
#include "../include/Stackgen.h"
#include "../include/buildstmt.h"
#include "../include/buildsymutil.h"
#include "../include/buildzplstmt.h"
#include "../include/createvar.h"
#include "../include/datatype.h"
#include "../include/db.h"
#include "../include/dimension.h"
#include "../include/error.h"
#include "../include/expr.h"
#include "../include/global.h"
#include "../include/rmstack.h"
#include "../include/runtime.h"
#include "../include/symbol.h"
#include "../include/symboltable.h"
#include "../include/typeinfo.h"


/* globals used to minimize the amount of parameter-passing necessary */
/* most of these store frequently-used symbol table values.           */

static void gen_glob_decls(FILE *outfile,symboltable_t* reg) {
  priv_tid_decl(outfile);
  /*** tid: required for any routine that accesses global data ***/
  priv_reg_alloc(outfile, reg);
  fprintf(outfile,"\n");
}


static void gen_loc_decls(FILE *outfile, char* regname, datatype_t* regdt) {
  if (!datatype_is_dense_reg(regdt)) {
    gen_sps_loc_decls(outfile,regname, regdt);
  }
    
  priv_tid_decl(outfile);
  /*** tid: required for any routine that accesses global data ***/
  fprintf(outfile,"\n");
}


static void gen_adjust(FILE* outfile, int global, char* regname, 
		       expr_t* basereg, expr_t* dir, expr_t* initexpr) {
  fprintf(outfile,"_AdjustRegionWith");
  switch (T_SUBTYPE(initexpr)) {
  case IN_REGION:
    fprintf(outfile,"In");
    break;
  case OF_REGION:
    fprintf(outfile,"Of");
    break;
  case AT_REGION:
    fprintf(outfile,"At");
    break;
  case BY_REGION:
    fprintf(outfile,"By");
    break;
  default:
    INT_FATAL(NULL,"Unknown region type in gen_glob_prep_reg_info");
  }
  if (global) {
    fprintf(outfile,"Glob");
  } else {
    fprintf(outfile, "Loc");
  }
  fprintf(outfile, "(");

  fprintf(outfile, regname);
  fprintf(outfile,",");
  gen_name(outfile, basereg);
  fprintf(outfile,",");
  gen_expr(outfile, dir);
  fprintf(outfile,");\n");
}


static void gen_glob_prep_reg_info(FILE *outfile, char* regname, 
				   expr_t* initexpr) {
  expr_t* basereg;
  expr_t* dir;
  symboltable_t* dirpst;

  basereg = T_OPLS(initexpr);

  dir = T_NEXT(T_OPLS(initexpr));
  typeinfo_expr(dir);
  dirpst = T_IDENT(dir);
  if (dirpst && (S_SETUP(dirpst) != 1)) {
    force_pst_init(outfile, dirpst);
  }
  gen_adjust(outfile, 1, regname, basereg, dir, initexpr);
}


static void gen_reg_bound(FILE *outfile,expr_t *bound,int dim) {
  if (bound == NULL) {
    fprintf(outfile,"?");
  } else {
    gen_expr(outfile, bound);
  }
}


static void gen_lo_bound(FILE *outfile,dimension_t *dimptr,int dim) {
  expr_t *dimexpr;
  
  if (dimptr == NULL) {
    dimexpr = NULL;
  } else {
    dimexpr = DIM_LO(dimptr);
  }
  typeinfo_expr(dimexpr);
  gen_reg_bound(outfile,dimexpr,dim);
}


static void gen_hi_bound(FILE *outfile,dimension_t *dimptr,int dim) {
  expr_t *dimexpr;
  
  if (dimptr == NULL) {
    dimexpr = NULL;
  } else {
    dimexpr = DIM_HI(dimptr);
  }
  typeinfo_expr(dimexpr);
  gen_reg_bound(outfile,dimexpr,dim);
}


/* This routine generates code to setup a region's global information */

static int gen_glob_norm_reg_info(FILE *outfile, symboltable_t* reg,
				  char* regname, datatype_t* regdt, 
				  dimension_t* dimptr) {
  int i;
  int flooddim;
  int rgriddim;
  int numdims = D_REG_NUM(regdt);
  int retval = 1;

  if (use_default_domain) {
    fprintf(outfile, "_REG_DOM(");
    fprintf(outfile, regname);
    fprintf(outfile, ") = &_DefaultDomain;\n");
    /*    fprintf(outfile, "_DOM_REGISTER(&_DefaultDomain, %s);\n", regname);*/
  }
  else {
    if (D_REG_DOM(regdt) == NULL) {
      if (!RMSCurrentDomain()) {
	fprintf(outfile, "_REG_DOM(%s) = &_DefaultDomain;\n", regname);
	/*	fprintf(outfile, "_DOM_REGISTER(&_DefaultDomain, %s);\n", regname);*/
      }
      else {
	fgenf(outfile, "_REG_DOM(%s) = %E;\n", regname, RMSCurrentDomain());
	/*	fgenf(outfile, "_DOM_REGISTER(%E, %s);\n", RMSCurrentDomain(), regname);*/
      }
    } else {
      fprintf(outfile, "_REG_DOM(%s) = &%s;\n", regname, S_IDENT(D_REG_DOM(regdt)));
      /*      fprintf(outfile, "_DOM_REGISTER(&%s, %s);\n", S_IDENT(D_REG_DOM(regdt)), regname);*/
    }
  }
  fprintf(outfile,"_NUMDIMS(");
  fprintf(outfile, regname);
  fprintf(outfile,") = %d;\n",numdims);
  fgenf(outfile, "_REG_ARRLIST(%s) = NULL;\n", regname);

  if (dimptr == NULL) {
    fprintf(outfile, "/* NULL dimptr */\n");
    return 0;
  }
  for (i=0;i<numdims;i++) {
    if (DIM_TYPE(dimptr) == DIM_INHERIT) {
      fprintf(outfile, "_INHERIT_GLOB_BOUNDS(");
      fprintf(outfile, regname);
      fprintf(outfile, ",%s.%s,%d);\n", RMS, RMS_REG, i);
    } else if (DIM_TYPE(dimptr) == DIM_DYNAMIC) {
      fprintf(outfile, "/* dimension %d dynamic */\n", i);
      retval = 0;
    } else {
      rgriddim = dimension_rgrid(dimptr);
      flooddim = dimension_floodable(dimptr) || rgriddim;
    
      fprintf(outfile,"_SET_GLOB_BOUNDS");
      if (flooddim) {
	fprintf(outfile,"_FLOOD");
      }
      fprintf(outfile,"(");
      fprintf(outfile, regname);
      fprintf(outfile,",%d,",i);
      if (flooddim) {
	fprintf(outfile,"0,0,_DIM_");
	if (rgriddim) {
	  fprintf(outfile,"RGRID");
	} else {
	  fprintf(outfile,"FLOODED");
	}
      } else {
	gen_lo_bound(outfile,dimptr,i);
	fprintf(outfile,",");
	gen_hi_bound(outfile,dimptr,i);
	fprintf(outfile,",1");
      }
      fprintf(outfile,");\n");
    }
    dimptr=DIM_NEXT(dimptr);
  }
  return retval;
}


static int gen_glob_reg_info(FILE *outfile, char* regname, 
			      symboltable_t* reg, datatype_t* regdt, 
			      expr_t* initexpr) {
  int retval = 0;

  switch (T_TYPE(initexpr)) {
  case SBE:
    retval = gen_glob_norm_reg_info(outfile, reg, regname, regdt,
				    T_SBE(initexpr));
    break;
  case BIPREP:
    gen_glob_prep_reg_info(outfile, regname, initexpr);
    retval = 1;
    break;
  case CONSTANT:
  case VARIABLE:
    retval = gen_glob_reg_info(outfile, regname, reg, regdt,
			       IN_VAL(S_INIT(T_IDENT(initexpr))));
    break;
  case BIWITH:
    typeinfo_expr(T_OPLS(initexpr));
    retval = gen_glob_reg_info(outfile, regname, reg, 
			       T_TYPEINFO(T_OPLS(initexpr)),
			       T_OPLS(initexpr));
    if (!datatype_is_dense_reg(regdt)) {
      fprintf(outfile," _SPS_MEMREQ(%s) = _%s_spsmemreqs;\n", regname, 
	      S_IDENT(reg));
    }
    break;
  default:
    INT_FATAL(NULL, "Unexpected case in gen_glob_reg_info");
  }

  return retval;
}


static void gen_expand_bounds(FILE *outfile,char* regname, 
			      datatype_t* regdt) {
  int numdims = D_REG_NUM(regdt);

  fprintf(outfile,"_EXPAND_DOM_BOUNDS(%s, %d);\n",regname, numdims);
}


static void gen_sparse_setup(FILE* outfile, char* regname, 
			     datatype_t* regdt) {
  fprintf(outfile,"_SPS_REGION(");
  fprintf(outfile, regname);
  fprintf(outfile,") = %d;\n",(datatype_is_dense_reg(regdt) == 0));
}


static void gen_basic_setup(FILE* outfile, char* regname) {
  fprintf(outfile,"_SetupRegion(");
  fprintf(outfile, regname);
  fprintf(outfile,");\n");
}


void GenGlobalSetup(FILE *outfile, 
		    symboltable_t* reg, datatype_t*regdt,
		    initial_t* init, char* regname) {
  expr_t* initexpr = IN_VAL(init);
  int expand;

  if (datatype_reg_dynamicdim(regdt)) {
    return;
  }

  if (T_TYPE(initexpr) == BIPREP) {
    expr_t* basereg;
    symboltable_t* baseregpst;
    
    basereg = T_OPLS(initexpr);
    baseregpst = T_IDENT(basereg);
    if (baseregpst && (S_SETUP(baseregpst) == 0)) {
      force_pst_init(outfile, baseregpst);
    }
  }

  fprintf(outfile, "{\n");
  gen_glob_decls(outfile,reg);

  gen_sparse_setup(outfile, regname, regdt);

  /* generate global stuff */
  expand = gen_glob_reg_info(outfile, regname, reg, regdt, IN_VAL(init));

  if (expand) {
    /* expand propagated bounds as necessary */
    gen_expand_bounds(outfile,regname, regdt);
  }

  fprintf(outfile,"}\n\n");
}


void GenLocalSetup(FILE *outfile, symboltable_t* reg,
		   datatype_t* regdt, initial_t* init, char* regname) {

  expr_t* initexpr = IN_VAL(init);

  if (datatype_reg_dynamicdim(regdt)) {
    return;
  }
			  
  if (T_TYPE(initexpr) == BIPREP) {
    expr_t* basereg;
    symboltable_t* baseregpst;
    
    basereg = T_OPLS(initexpr);
    baseregpst = T_IDENT(basereg);
    if (baseregpst && (S_SETUP(baseregpst) == 0)) {
      force_pst_finalize(outfile, baseregpst);
    }
  }
  fprintf(outfile, "{\n");
  gen_loc_decls(outfile, regname, regdt);

  fprintf(outfile,"_REG_LABEL(%s) = \"%s\";\n", regname, regname);


  gen_basic_setup(outfile, regname);
  if (T_TYPE(initexpr) == BIPREP) {
    gen_adjust(outfile, 0, regname, T_OPLS(initexpr), 
	       T_NEXT(T_OPLS(initexpr)), initexpr);
  }


  if (!datatype_is_dense_reg(regdt)) {
    gen_sparse_fluff(outfile, regname, reg, regdt);
    gen_sparse_structure(outfile, reg, regname, regdt, initexpr);
  }

  fprintf(outfile,"}\n\n");
}

