/******************************************************************************

            Copyright (c) 1996 - 2004  --  University of Washington

******************************************************************************/


#include <stdio.h>
#include <string.h>
#include "../include/Cgen.h"
#include "../include/Dgen.h"
#include "../include/Privgen.h"
#include "../include/Stackgen.h"
#include "../include/buildstmt.h"
#include "../include/buildzplstmt.h"
#include "../include/db.h"
#include "../include/dimension.h"
#include "../include/error.h"
#include "../include/expr.h"
#include "../include/glist.h"
#include "../include/glistmac.h"
#include "../include/global.h"
#include "../include/runtime.h"
#include "../include/symboltable.h"
#include "../include/datatype.h"

static void GenUpDownFluff(FILE *outfile,char* arrname,int strided,
			   int fluffval,int mfluff,int wrapfluff,
			   int dim, int lohi) {
  char regname[1024];
  
  sprintf(regname, "_ARR_DECL_REG(%s)", arrname);
  gen_fluffval(outfile,fluffval,mfluff,dim, regname);
}

static void DetermineRegionForEnsemble(symboltable_t* pst,
				       datatype_t*pdt,
				       expr_t** reg) {
  *reg = symtab_find_reg(pst);
  if (*reg == NULL) {
    *reg = buildqregexpr(D_ENS_NUM(pdt));
  }
  if (datatype_reg_dynamicdim(T_TYPEINFO(*reg))) {
    *reg = buildqregexpr(D_REG_NUM(T_TYPEINFO(*reg)));
  }
}


void GenAllocEnsemble(FILE* outfile, symboltable_t* pst, char* arrayname, 
		      datatype_t* pdt, int isrealloc, int hoistable) {
  expr_t* reg;
  int numdims;
  int strided;
  int wrap;
  int arrtype;
  char* accname;
  int i;
  int dynamic;
  int push_scope;
  region_t dyn_reg_scope;

  DetermineRegionForEnsemble(pst, pdt, &reg);
  numdims = D_REG_NUM(T_TYPEINFO(reg));
  dynamic = (expr_dyn_reg(reg) || 
	     (expr_contains_qreg(reg) && !expr_is_qreg(reg)));
  push_scope = (dynamic && S_IDENT(pst)[0] != '_');
  if (expr_is_qreg(reg)) {
    push_scope = 0;
  }
  if (push_scope) {
    dyn_reg_scope.region = reg;
    dyn_reg_scope.mask = NULL;
    dyn_reg_scope.with = MASK_NONE;
    dyn_reg_scope.body_p = NULL;
    fprintf(outfile, "{\n");
    MakeDRDeclarations(outfile, T_REGION_SYM(&dyn_reg_scope));
    fprintf(outfile, "\n");
    SetupDRs(outfile, T_REGION_SYM(&dyn_reg_scope), NULL);
  }

  strided = expr_is_strided_reg(reg, 1);
  wrap = 0;
  for (i = 0; i < numdims; i++) {
    wrap = wrap || S_WRAPFLUFF_LO(pst,i) || S_WRAPFLUFF_HI(pst,i);
  }
  arrtype = ARR_NORMAL;
  if (expr_is_strided_reg(reg,1)) {
    arrtype = ARR_STRIDED;
  }
  if (!expr_is_dense_reg(reg)) {
    arrtype = ARR_SPARSE;
  }
  accname = gen_accessors(pst, reg, numdims, arrtype, pdt);

  /* fluff stuff */
  fgenf(outfile, "{\n");
  fgenf(outfile, "int _down[%d];\n", numdims);
  fgenf(outfile, "int _up[%d];\n", numdims);
  fprintf(outfile, "_ARR_DECL_REG(%s) = ", arrayname);
  priv_reg_access(outfile,reg);
  fprintf(outfile, ";\n");
  for (i=0; i < numdims; i++) {
    fgenf(outfile, "_down[%d] = ", i);
    GenUpDownFluff(outfile, arrayname, strided, S_FLUFF_LO(pst, i),
		   S_UNK_FLUFF_LO(pst,i), S_WRAPFLUFF_LO(pst,i), i, LO);
    fgenf(outfile, ";\n");
  }
  for (i=0; i < numdims; i++) {
    fgenf(outfile, "_up[%d] = ", i);
    GenUpDownFluff(outfile, arrayname, strided, S_FLUFF_HI(pst, i),
		   S_UNK_FLUFF_HI(pst,i), S_WRAPFLUFF_HI(pst,i), i, HI);
    fgenf(outfile, ";\n");
  }

  fgenf(outfile, "_ZPL_ALLOC_ARRAY(");
  fgenf(outfile, "%s, ", arrayname);
  priv_reg_access(outfile, reg);
  fprintf(outfile, ", ");
  if (D_CLASS(D_ENS_TYPE(pdt)) == DT_ARRAY) {
    dimension_t* pdTemp = D_ARR_DIM(D_ENS_TYPE(pdt));
    if (!expr_computable_const(DIM_LO(pdTemp)) ||
	!expr_computable_const(DIM_HI(pdTemp))) {
      fgenf(outfile, "sizeof(%D) * ((%E)-(%E)+1), ", D_ARR_TYPE(D_ENS_TYPE(pdt)), PDT_SIZE,
	    DIM_HI(pdTemp), DIM_LO(pdTemp));
    }
    else {
      fgenf(outfile, "sizeof(%D), ", D_ENS_TYPE(pdt), PDT_SIZE);
    }
  }
  else {
    fgenf(outfile, "sizeof(%D), ", D_ENS_TYPE(pdt), PDT_SIZE);
  }
  fgenf(outfile, "_up, _down, %d, %d, ", wrap, arrtype == ARR_SPARSE);
  fgenf(outfile, "%s, \"%s\", %d, %d);\n", accname, arrayname, isrealloc, D_ENS_DNR(pdt));
  if (hoist_access_ptrs == HOIST_PTRS_TO_SUBPROG && 
      hoistable) {  /* we can't handle indarrays or structures here yet */
    dimension_t* arrdim;

    fprintf(outfile, "_INIT_HOISTED_PTR(");
    arrdim = gen_pdt(outfile, D_ENS_TYPE(pdt), PDT_PCST);
    if (arrdim) {
      fprintf(outfile, "*");
    }
    fprintf(outfile, ", %s", arrayname);
    fprintf(outfile, ");\n");
  }

  fprintf(outfile, "}\n");

  if (push_scope) {
    fprintf(outfile, "}\n");
  }
}


static void GenAllocEnsembleRec(FILE* f,
				symboltable_t* pst,
				datatype_t* pdt,
				char* arrayname,
				int isrealloc) {
  static int calldepth=0;
  expr_t* reg;
  int i;
  int numdims;
  dimension_t *pdim;
  dimension_t *pdTemp;
  symboltable_t* pstTemp;
  int depth = 0;
  region_t dyn_reg_scope;
  int dynamic;
  int push_scope;
  symboltable_t* indpst[MAXRANK];

  calldepth++;

  if (D_CLASS(pdt) == DT_ARRAY) {
    pdim = D_ARR_DIM(pdt);
    fprintf(f, "{\n"); 
    depth++;
    for (i=0;i<D_ARR_NUM(pdt);i++) {
      indpst[i] = get_blank_arrayref_index(calldepth, i+1);
      fprintf(f, "int %s;\n", S_IDENT(indpst[i]));
    }
    fprintf(f,"\n");
    pdTemp=pdim;
    for (i=0; i<D_ARR_NUM(pdt); i++) {
      fprintf(f, "for (%s=", S_IDENT(indpst[i]));
      gen_expr(f, DIM_LO(pdTemp));
      fprintf(f, "; %s <= (", S_IDENT(indpst[i]));
      gen_expr(f, DIM_HI(pdTemp));
      fprintf(f, "); %s++) {\n", S_IDENT(indpst[i]));
      depth++;

      if (expr_computable_const(DIM_LO(pdTemp))) {
	int lo = expr_intval(DIM_LO(pdTemp));
	if (lo == 0) {
	  sprintf(arrayname, "%s[%s]", arrayname, S_IDENT(indpst[i]));
	} else {
	  sprintf(arrayname, "%s[%s-(%d)]", arrayname, S_IDENT(indpst[i]), lo);
	}
      } else {
	INT_FATAL(T_STMT(DIM_LO(pdTemp)), "Seems I can only handle arrays "
		  "with constant lower bounds, currently.");
      }
      pdTemp=DIM_NEXT(pdTemp);
    }
    GenAllocEnsembleRec(f, pst, D_ARR_TYPE(pdt), arrayname, isrealloc);
    while (depth-- > 0) {
      fprintf(f, "}\n");
    }
    calldepth--;
    return;
  }

  if (D_CLASS(pdt) == DT_STRUCTURE) {
    pstTemp = D_STRUCT(pdt);
    while (pstTemp!=NULL) {
      if (datatype_find_ensemble(S_DTYPE(pstTemp))) {
	char newarrayname[256];

	sprintf(newarrayname, "%s.%s", arrayname, S_IDENT(pstTemp));
	GenAllocEnsembleRec(f, pstTemp, S_DTYPE(pstTemp), newarrayname, isrealloc);
      }
      pstTemp = S_SIBLING(pstTemp);
    }
    calldepth--;
    return;
  }

  if (D_CLASS(pdt) != DT_ENSEMBLE) {
    INT_FATAL(NULL, "Ensemble type expected in GenAllocEnsembleRec\n");
  }

  DetermineRegionForEnsemble(pst, pdt, &reg);

  numdims = D_REG_NUM(T_TYPEINFO(reg));
  dynamic = (expr_dyn_reg(reg) || 
	     (expr_contains_qreg(reg) && !expr_is_qreg(reg)));
  push_scope = (dynamic && S_IDENT(pst)[0] != '_');
  if (expr_is_qreg(reg)) {
    push_scope = 0;
  }
  if (push_scope) {
    dyn_reg_scope.region = reg;
    dyn_reg_scope.mask = NULL;
    dyn_reg_scope.with = MASK_NONE;
    dyn_reg_scope.body_p = NULL;
    fprintf(f, "{\n"); depth++;
    MakeDRDeclarations(f, T_REGION_SYM(&dyn_reg_scope));
    fprintf(f, "\n");
    SetupDRs(f, T_REGION_SYM(&dyn_reg_scope), NULL);
  }
	  
  if (strcmp(arrayname, S_IDENT(pst)) != 0) {
    fprintf(f, "%s = &(_E_%s);\n", arrayname, arrayname);
  }

  GenAllocEnsemble(f, pst, arrayname, pdt, isrealloc, calldepth == 1);

  while (depth-- > 0) {
    fprintf(f, "}\n");
  }

  calldepth--;
}

static void GenFreeEnsembleRec(FILE* f,
			       symboltable_t* pst,
			       datatype_t* pdt,
			       char* arrayname) {
  static int calldepth=0;
  expr_t* reg;
  int i;
  int strided;
  int numdims;
  int arrtype;
  char* accname;
  int wrap;
  dimension_t *pdim;
  dimension_t *pdTemp;
  symboltable_t* pstTemp;
  int depth = 0;
  symboltable_t* indpst[MAXRANK];

  calldepth++;

  if (D_CLASS(pdt) == DT_ARRAY) {
    pdim = D_ARR_DIM(pdt);
    fprintf(f, "{\n");
    depth++;
    for (i=0;i<D_ARR_NUM(pdt);i++) {
      indpst[i] = get_blank_arrayref_index(calldepth, i+1);
      fprintf(f, "int %s;\n", S_IDENT(indpst[i]));
    }
    fprintf(f,"\n");
    pdTemp=pdim;
    for (i=0; i<D_ARR_NUM(pdt); i++) {
      fprintf(f, "for (%s=0; %s < ((", S_IDENT(indpst[i]), S_IDENT(indpst[i]));
      gen_expr(f, DIM_HI(pdTemp));
      fprintf(f,") - (");
      gen_expr(f, DIM_LO(pdTemp));
      fprintf(f,") + 1");
      fprintf(f,"); %s++) {\n", S_IDENT(indpst[i]));
      depth++;

      pdTemp=DIM_NEXT(pdTemp);
      sprintf(arrayname, "%s[%s]", arrayname, S_IDENT(indpst[i]));
    }
    GenFreeEnsembleRec(f, pst, D_ARR_TYPE(pdt), arrayname);
    while (depth-- > 0) {
      fprintf(f, "}\n");
    }
    calldepth--;
    return;
  }

  if (D_CLASS(pdt) == DT_STRUCTURE) {
    pstTemp = D_STRUCT(pdt);
    while (pstTemp!=NULL) {
      if (datatype_find_ensemble(S_DTYPE(pstTemp))) {
        char newarrayname[256];

        sprintf(newarrayname, "%s.%s", arrayname, S_IDENT(pstTemp));
        GenFreeEnsembleRec(f, pstTemp, S_DTYPE(pstTemp), newarrayname);
      }
      pstTemp = S_SIBLING(pstTemp);
    }
    calldepth--;
    return;
  }

  if (D_CLASS(pdt) != DT_ENSEMBLE) {
    INT_FATAL(NULL, "Ensemble type expected in GenFreeEnsembleRec\n");
  }

  DetermineRegionForEnsemble(pst, pdt, &reg);

  strided = expr_is_strided_reg(reg, 1);
  numdims = D_REG_NUM(T_TYPEINFO(reg));
  wrap = 0;
  for (i = 0; i < numdims; i++) {
    wrap = wrap || S_WRAPFLUFF_LO(pst,i) || S_WRAPFLUFF_HI(pst,i);
  }

  arrtype = ARR_NORMAL;
  if (expr_is_strided_reg(reg,1)) {
    arrtype = ARR_STRIDED;
  }
  if (!expr_is_dense_reg(reg)) {
    arrtype = ARR_SPARSE;
  }
  accname = gen_accessors(pst, reg, numdims, arrtype, pdt);

  fgenf(f, "_ZPL_FREE_ARRAY(%s,%d);\n", arrayname, D_ENS_DNR(pdt));

  while (depth-- > 0) {
    fprintf(f, "}\n");
  }

  calldepth--;
}

void GenAllocEnsembleOld(FILE *f, symboltable_t *pst) {
  datatype_t* pdt;
  char arrayname[256] = "";

  pdt = S_DTYPE(pst);
  if (datatype_find_ensemble(pdt)) {
    sprintf(arrayname, "%s", S_IDENT(pst));
    GenAllocEnsembleRec(f, pst, pdt, arrayname, 0);
  }
}

void GenReallocEnsemble(FILE *f, symboltable_t *pst) {
  datatype_t* pdt;
  char arrayname[256] = "";

  pdt = S_DTYPE(pst);
  if (datatype_find_ensemble(pdt)) {
    sprintf(arrayname, "%s", S_IDENT(pst));
    GenAllocEnsembleRec(f, pst, pdt, arrayname, 1);
  }
}

void GenFreeEnsemble(FILE *f, symboltable_t *pst) {
  datatype_t* pdt;
  char arrayname[256] = "";

  pdt = S_DTYPE(pst);
  if (datatype_find_ensemble(pdt)) {
    sprintf(arrayname, "%s", S_IDENT(pst));
    GenFreeEnsembleRec(f, pst, pdt, arrayname);
  }
}
