/******************************************************************************

            Copyright (c) 1996 - 2004  --  University of Washington

******************************************************************************/

#ifndef _DOMAIN_H
#define _DOMAIN_H

#include "struct.h"
#include "zplglobal.h"
#include "type.h"
#include "grid.h"
#include "ironman.h"

#define _PROC_TO_DATA(x)   ((int)((x)+(((double)(x))!=((double)((int)(x))))))
#define _DATA_TO_PROC(x)   ((int)(x))

#define _BND_LO(bnd) ((bnd).lo)
#define _BND_HI(bnd) ((bnd).hi)

#define _DOM_GRD(dom) ((dom).grid)
#define _DOM_GRDSIZE_V(dom) (_GRD_SIZE_V(*(_DOM_GRD(dom))))
#define _DOM_GRDSIZE(dom, dim) (_GRD_SIZE(*(_DOM_GRD(dom)), dim))
#define _DOM_GRDLOC_V(dom) (_GRD_LOC_V(*(_DOM_GRD(dom))))
#define _DOM_GRDLOC(dom, dim) (_GRD_LOC(*(_DOM_GRD(dom)), dim))
#define _DOM_GRDBLK_V(dom) (_GRD_BLK_V(*(_DOM_GRD(dom))))
#define _DOM_GRDBLK(dom, dim) (_GRD_BLK(*(_DOM_GRD(dom)), dim))
#define _DOM_DIS(dom, dim) ((dom).dis[dim])
#define _DIS_LOCAL_LO(dis) (_BND_LO((dis).localBound))
#define _DIS_LOCAL_HI(dis) (_BND_HI((dis).localBound))
#define _DIS_TO_PROC(dis) ((dis).toProc)
#define _DIS_GLOBAL_LO(dis) (_BND_LO((dis).globalBound))
#define _DIS_GLOBAL_HI(dis) (_BND_HI((dis).globalBound))
#define _DIS_GLOBAL_SIZE(dis) (_DIS_GLOBAL_HI(dis) - _DIS_GLOBAL_LO(dis) + 1)

#define _DOM_MYLO(dom, dim) (_DIS_LOCAL_LO(*_DOM_DIS(dom, dim)))
#define _DOM_MYHI(dom, dim) (_DIS_LOCAL_HI(*_DOM_DIS(dom, dim)))
#define _DOM_LO(dom, dim) (_DIS_GLOBAL_LO(*_DOM_DIS(dom, dim)))
#define _DOM_HI(dom, dim) (_DIS_GLOBAL_HI(*_DOM_DIS(dom, dim)))
#define _DOM_SIZE(dom, dim) (_DIS_GLOBAL_SIZE(*_DOM_DIS(dom, dim)))

#define _DOM_LABEL(dom) ((dom).domain_label)

#define _DOM_REGISTER(dom, reg) _domain_register_region(dom, reg)
#define _DOM_UNREGISTER(dom, reg) _domain_unregister_region(dom, reg)
#define _DOM_REGLIST(dom) ((dom).regions)
#define _DOM_REGLIST_REG(drlist) ((drlist)->reg)
#define _DOM_REGLIST_NEXT(drlist) ((drlist)->next)

/** name is blocked, ... unknown if unknown **/
#define _DOM_TO_PROC(name, dom, i, dim) (_fast_##name##ToProc(dom, dim, i))
#define _DOM_TO_PROC_BLK(name, dom, i, dim) (_DOM_TO_PROC(name, dom, i, dim) * _DOM_GRDBLK(dom, dim))

#define _fast_unknownToProc(dom, dim, i) (_DIS_TO_PROC(*_DOM_DIS(dom, dim))(&dom, dim, i))

#define _MAXMINIMIZE_DOM(dom) \
  { \
    int i; \
    for (i=0; i<_MAXRANK; i++) { \
      _BND_LO(_invisibleBounds[i]) = INT_MAX; \
      _BND_HI(_invisibleBounds[i]) = INT_MIN; \
    } \
  }

typedef struct __bound {
  int lo; /* lower bound */
  int hi; /* upper bound */
} _bound;

struct __distribution_info {
  _bound localBound;
  _bound globalBound;
  int (*toProc)(_domain, int, int);
};

struct __region_list {
  _region_nc reg;
  _region_list* next;
};

struct __domain_info {
  _grid_nc grid;                     /* domain's grid          */
  _region_list* regions;             /* domain's regions       */
  _distribution_info* dis[_MAXRANK]; /* domain's distributions */
  char* domain_label;                /* domain label           */
  _MD_DOMAIN_INFO                    /* machine dependent info */
};

_domain_info _DefaultDomain;

extern _bound _invisibleBounds[_MAXRANK];


void _InitDomain(_domain_fnc, _grid_fnc, char*);
void _domain_register_region(_domain_fnc, _region_fnc);
void _domain_unregister_region(_domain_fnc, _region);
void _print_domain_hierarchy(_domain);

#define _InitDomainDist(dom, dim, dist) \
  _DOM_DIS(*dom, dim) = _zmalloc(sizeof(_blocked_info), "domain distribution")

/*  _##dist##Setup(dom, dim, _BND_LO(_invisibleBounds[dim]), _BND_HI(_invisibleBounds[dim]))*/


/*** Blocked Distribution ***/

typedef struct {
  _bound localBound;
  _bound globalBound;
  int (*toProc)(_domain, int, int);
} _blocked_info;
typedef _blocked_info* _blocked;

void _blockedSetup(_domain_fnc dom, int dim, int lo, int hi);
int _blockedToProc(_domain dom, int dim, int i);

#define _fast_blockedToProc(dom, dim, i) \
  (_DATA_TO_PROC((((double)(i) - _DOM_LO(dom, dim)) * _DOM_GRDSIZE(dom, dim)) / _DOM_SIZE(dom, dim)))

/*** Goofer Distribution ***/

typedef struct {
  _bound localBound;
  _bound globalBound;
  int (*toProc)(_domain, int, int);
} _goofer_info;
typedef _goofer_info* _goofer;

void _gooferSetup(_domain_fnc dom, int dim);
int _gooferToProc(_domain dom, int dim, int i);

#define _fast_gooferToProc(dom, dim, i) \
  (_gooferToProc(&dom, dim, i))

/*** Fourfer Distribution ***/

typedef struct {
  _bound localBound;
  _bound globalBound;
  int (*toProc)(_domain, int, int);
} _fourfer_info;
typedef _fourfer_info* _fourfer;

void _fourferSetup(_domain_fnc dom, int dim);
int _fourferToProc(_domain dom, int dim, int i);

#define _fast_fourferToProc(dom, dim, i) \
  (_fourferToProc(&dom, dim, i))

#endif
