/******************************************************************************

            Copyright (c) 1996 - 2004  --  University of Washington

******************************************************************************/

#include <stdio.h>
#include "md_zinc.h"
#include "grid.h"
#include "domain.h"
#include "md_zlib.h"
#include "zlib.h"
#include "zplglobal.h"
#include "file_io.h"
#include "zerror.h"

int _GRD_TO_PROC(_grid grd, const int nd, const int loc[]) {
  int index;
  int i;

  index = 0;
  for (i=0; i<nd; i++) {
    index += loc[i]*_GRD_BLK(*grd, i);
  }
  return index;
}


int _GetSliceColor(_grid grd, int slicenum) {
  int i;
  int dimbit;
  int loc[_MAXRANK];

  if (slicenum == 0) {
    return 0;
  }

  dimbit = 1;
  for (i=0; i<_MAXRANK; i++) {
    if (dimbit & slicenum) {
      loc[i] = _GRD_LOC(*grd, i);
    } else {
      loc[i] = 0;
    }
    dimbit = dimbit << 1;
  }

  return _GRD_TO_PROC(grd, _MAXRANK,loc);
}


int _BegOfSlice(_grid grd, int slicenum) {
  int dimbit;
  int i;
  int first = 1;

  dimbit = 1;
  for (i=0; i<_MAXDIM_GRID; i++) {
    if (dimbit & slicenum) {
    } else {
      if (!_GRD_BEG(*grd, i)) {
	first = 0;
      }
    }
    dimbit = dimbit << 1;
  }
  return first;
}


int _EndOfSlice(_grid grd, int slicenum) {
  int dimbit;
  int i;
  int last = 1;

  dimbit = 1;
  for (i=0; i<_MAXDIM_GRID; i++) {
    if (dimbit & slicenum) {
    } else {
      if (!_GRD_END(*grd, i)) {
	last = 0;
      }
    }
    dimbit = dimbit << 1;
  }
  return last;
}


int _NextProcSameSlice(_grid grd, int slicenum) {
  int dimbit;
  int i;
  int loc[_MAXRANK];
  int carry = 1;

  dimbit = 1 << (_MAXDIM_GRID-1);
  for (i=_MAXDIM_GRID-1; i>=0; i--) {
    if (dimbit & slicenum) {
      loc[i] = _GRD_LOC(*grd, i);
    } else {
      if (carry) {
	if (_GRD_END(*grd, i)) {
	  loc[i] = 0;
	} else {
	  loc[i] = _GRD_LOC(*grd, i)+1;
	  carry = 0;
	}
      } else {
	loc[i] = _GRD_LOC(*grd, i);
      }
    }

    dimbit = dimbit >> 1;
  }

  return _GRD_TO_PROC(grd, _MAXDIM_GRID,loc);
}


int _PosToProcInSlice(_grid grd, const int numdims,const int loc[],const int slicenum) {
  int dimbit;
  int i;
  int ind;
  int procs;

  if (slicenum == 0) {
    return _GRD_TO_PROC(grd, numdims,loc);
  } else {
    ind = 0;
    procs = 1;
    dimbit = 1 << (_MAXDIM_GRID-1);
    for (i=_MAXDIM_GRID-1; i>=0; i--) {
      if (dimbit & slicenum) {
      } else {
	ind += loc[i]*procs;
	procs *= _GRD_SIZE(*grd, i);
      }
      dimbit = dimbit >> 1;
    }
   
    return ind;
  }
}


int _MaxProcInSlice(_grid grd, const int slicenum) {
  int dimbit;
  int i;
  int ind;

  if (slicenum == 0) {
    return _PROCESSORS-1;
  } else {
    ind = 1;
    dimbit = 1 << (_MAXDIM_GRID-1);
    for (i=_MAXDIM_GRID-1; i>=0; i--) {
      if (dimbit & slicenum) {
      } else {
	ind *= _GRD_SIZE(*grd, i);
      }

      dimbit = dimbit >> 1;
    }
    return ind-1;
  }
}


int _MyLocInSlice(_grid grd, int slicenum) {
  return _PosToProcInSlice(grd, _MAXDIM_GRID, _GRD_LOC_V(*grd), slicenum);
}


static int _GetNumSlices(void) {
  int i;
  int numslices;

  numslices = 1;
  for (i=0; i<_MAXDIM_GRID; i++) {
    numslices = numslices << 1;
  }
  numslices--;

  return numslices;
}


static void _InitSlices(_grid_fnc grd) {
  int i;
  int dim;
  int dimbit;
  int distributed;

  _GRD_NUMSLICES(*grd) = _GetNumSlices();
  _GRD_DISTRIBUTED_V(*grd) = (int*)_zmalloc(_GRD_NUMSLICES(*grd)*sizeof(int),
					      "slice distributed vector");
  for (i=0; i<_GRD_NUMSLICES(*grd); i++) {
    dimbit = 1;
    distributed = 0;
    for (dim=0; dim<_MAXDIM_GRID; dim++) {
      if (i & dimbit) {
      } else {
	if (_GRD_SIZE(*grd, dim) > 1) {
	  distributed = 1;
	}
      }
      dimbit = dimbit << 1;
    }
    _GRD_DISTRIBUTED(*grd, i) = distributed;
  }
}

#define _MAXFACTORS 32

/*
 * factors remaining processors amongst unallocated dimensions
 */
static void factor(int val, int factor[], int numslots) {
  int initval;
  int numfactors;
  int i;
  int newfactor;
  int foundslot;

  if (val <= 1) {
    numfactors = 1;
    factor[0] = 1;
  } else {
    initval = val;
    numfactors = 0;
    for (i=2; i<=initval; i++) {
      while (val%i == 0) {
	factor[numfactors] = i;
	numfactors++;
	val/=i;
      }
    }
  }

  if (numfactors > _MAXFACTORS) {
    fprintf(stderr,"ERROR: Brad knows nothing about number theory\n");
    exit(1);
  }

  if (numfactors > numslots) {
    while (numfactors > numslots) {
      numfactors--;
      newfactor = factor[0] * factor[1];

      foundslot = 0;
      for (i=0; !foundslot && i<(numfactors-1); i++) {
	if (newfactor < factor[i+2]) {
	  factor[i] = newfactor;
	  foundslot = 1;
	} else {
	  factor[i] = factor[i+2];
	}
      }
      if (foundslot) {
	for ( ; i<numfactors; i++) {
	  factor[i] = factor[i+1];
	}
      } else {
	factor[numfactors-1] = newfactor;
      }
    }
  }

  if (numfactors < numslots) {
    while (numfactors < numslots) {
      numfactors++;
      factor[numfactors-1] = 1;
    }
  }
}

void _InitGrid(_grid_fnc grd, int init[], int rank, char* label) {
  int i;
  int allocated;
  int factors[_MAXFACTORS];
  int unallocated;
  int procs;
  int index;
  
  _GRD_LABEL(*grd) = label;

  allocated = 1;
  unallocated = 0;
  for (i = 0; i < rank; i++) {
    _GRD_SIZE(*grd, i) = init[i];
    if (init[i] == 0) {
      unallocated++;
    }
    else {
      allocated *= init[i];
    }
  }

  if (allocated > _PROCESSORS) {
    fprintf(stderr, "Allocated too many processors to grid\n");
    exit(1);
  }

  if (unallocated > 0) {
    factor(_PROCESSORS / allocated, factors, unallocated);
    for (i = 0; i < rank; i++) {
      if (_GRD_SIZE(*grd, i) == 0) {
	_GRD_SIZE(*grd, i) = factors[unallocated - 1];
	unallocated--;
      }
    }
  }

  if (unallocated != 0) {
    fprintf(stderr, "Did not allocate processors to grid\n");
    exit(1);
  }

  for (i = rank; i < _MAXRANK; i++) {
    _GRD_SIZE(*grd, i) = 1;
  }

  procs = _PROCESSORS;
  index = _INDEX;
  for (i = 0; i < _MAXRANK; i++) {
    procs /= _GRD_SIZE(*grd, i);
    _GRD_BLK(*grd, i) = procs;
    _GRD_LOC(*grd, i) = index / procs;
    index = index % procs;
  }
  /*
    for ( i = 0; i < _MAXRANK; i++) {
    printf("%d ", _GRD_SIZE(*grd, i));
    }
    printf("\n");
  */
  _InitSlices(grd);

  _MDInitGrid(grd);
  _InitAtCommPerGrid(grd);
}


void _InitNodeStateVector(void) {
  int i;
  int procs;
  int index;

  procs = _PROCESSORS;
  index = _INDEX;
  for (i = 0; i < _MAXRANK; i++) {
    procs /= _GRD_SIZE(_DefaultGrid, i);

    _GRD_BLK(_DefaultGrid, i) = procs;
    _GRD_LOC(_DefaultGrid, i) = index / procs;

    index = index%procs;
  }

  _GRD_LABEL(_DefaultGrid) = "Implicit Grid";

  _InitSlices(&_DefaultGrid);
  _MDInitGrid(&_DefaultGrid);
  _InitAtCommPerGrid(&_DefaultGrid);
}


void _PrintGridConfiguration(_grid grd) {
  int i;

  if (_INDEX == 0) {
    printf("Using %d processors\n",_PROCESSORS);
    if (_MAXDIM == 0) {
      printf("Computation is completely sequential\n");
    } else {
      if (_MAXDIM_GRID > 1) {
	printf("Using a %d",_GRD_SIZE(*grd, 0));
	for (i=1; i<_MAXDIM; i++) {
	  printf(" x %d",_GRD_SIZE(*grd, i));
	}
	if (i<_MAXDIM_GRID) {
	  printf(" (");
	  for ( ; i<_MAXDIM_GRID; i++) {
	    printf(" x %d",_GRD_SIZE(*grd, i));
	  }
	  printf(" wasted)");
	}
	printf(" processor grid\n");
      }
    }
    fflush(stdout);
  }
}

void _grid_register_domain(_grid_fnc grd, _domain_fnc dom) {
  _domain_list* tmp = _zmalloc(sizeof(_domain_list), "grid domain list node");

  _GRD_DOMLIST_DOM(tmp) = dom;
  _GRD_DOMLIST_NEXT(tmp) = _GRD_DOMLIST(*grd);
  _GRD_DOMLIST(*grd) = tmp;
}

void _grid_unregister_domain(_grid_fnc grd, _domain dom) {
  _domain_list* tmp;
  _domain_list* prev;

  prev = NULL;
  tmp = _GRD_DOMLIST(*grd);
  while (tmp) {
    if (_GRD_DOMLIST_DOM(tmp) == dom) {
      break;
    }
    prev = tmp;
    tmp = _GRD_DOMLIST_NEXT(tmp);
  }
  if (!tmp) {
    _RT_ANY_FATAL1("Domain %s cannot be unregistered", _DOM_LABEL(*dom));
  }
  if (prev) {
    _GRD_DOMLIST_NEXT(prev) = _GRD_DOMLIST_NEXT(tmp);
  }
  else {
    _GRD_DOMLIST(*grd) = _GRD_DOMLIST_NEXT(tmp);
  }
  _zfree(tmp, "grid domain list node");
}

void _print_grid_hierarchy(_grid grd) {
  _domain_list* tmp = _GRD_DOMLIST(*grd);

  _SCALAR_FPRINTF1(zout, "Grid %s\n", _GRD_LABEL(*grd));
  while (tmp) {
    _print_domain_hierarchy(_GRD_DOMLIST_DOM(tmp));
    tmp = _GRD_DOMLIST_NEXT(tmp);
  }
}
